@@ -68,7 +68,7 @@ int Unicon_fnSendKey (unsigned char ch, char flag)
    if (tty == NULL) {
        return -1;
    }
-   tty->ldisc.ops->receive_buf(tty, &ch, &mbz, 1);
+   tty->ldisc->ops->receive_buf(tty, &ch, &mbz, 1);
 #ifdef DEBUG
    printk ("will sending %c, 0x%x\n", ch, ch);
 #endif
