@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2005-2009 Junjiro R. Okajima
+ * Copyright (C) 2005-2011 Junjiro R. Okajima
  *
  * This program, aufs is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -33,7 +33,7 @@ static void au_h_nd(struct nameidata *h_nd, struct nameidata *nd)
 		 * due to whiteout and branch permission.
 		 */
 		h_nd->flags &= ~(/*LOOKUP_PARENT |*/ LOOKUP_OPEN | LOOKUP_CREATE
-				 | LOOKUP_FOLLOW);
+				 | LOOKUP_FOLLOW | LOOKUP_EXCL);
 		/* unnecessary? */
 		h_nd->intent.open.file = NULL;
 	} else
@@ -70,6 +70,7 @@ struct dentry *au_lkup_one(struct qstr *name, struct dentry *h_parent,
 		path_put(&h_nd.path);
 	}
 
+	AuTraceErrPtr(h_dentry);
 	return h_dentry;
 }
 
@@ -81,8 +82,10 @@ static void au_call_lkup_one(void *args)
 
 #define AuLkup_ALLOW_NEG	1
 #define au_ftest_lkup(flags, name)	((flags) & AuLkup_##name)
-#define au_fset_lkup(flags, name)	{ (flags) |= AuLkup_##name; }
-#define au_fclr_lkup(flags, name)	{ (flags) &= ~AuLkup_##name; }
+#define au_fset_lkup(flags, name) \
+	do { (flags) |= AuLkup_##name; } while (0)
+#define au_fclr_lkup(flags, name) \
+	do { (flags) &= ~AuLkup_##name; } while (0)
 
 struct au_do_lookup_args {
 	unsigned int		flags;
@@ -101,13 +104,11 @@ au_do_lookup(struct dentry *h_parent, struct dentry *dentry,
 {
 	struct dentry *h_dentry;
 	struct inode *h_inode, *inode;
-	struct qstr *name;
 	struct au_branch *br;
 	int wh_found, opq;
 	unsigned char wh_able;
 	const unsigned char allow_neg = !!au_ftest_lkup(args->flags, ALLOW_NEG);
 
-	name = &dentry->d_name;
 	wh_found = 0;
 	br = au_sbr(dentry->d_sb, bindex);
 	wh_able = !!au_br_whable(br->br_perm);
@@ -125,8 +126,8 @@ au_do_lookup(struct dentry *h_parent, struct dentry *dentry,
 	if (!allow_neg)
 		return NULL; /* success */
 
- real_lookup:
-	h_dentry = au_lkup_one(name, h_parent, br, args->nd);
+real_lookup:
+	h_dentry = au_lkup_one(&dentry->d_name, h_parent, br, args->nd);
 	if (IS_ERR(h_dentry))
 		goto out;
 
@@ -160,10 +161,10 @@ au_do_lookup(struct dentry *h_parent, struct dentry *dentry,
 	}
 	goto out;
 
- out_neg:
+out_neg:
 	dput(h_dentry);
 	h_dentry = NULL;
- out:
+out:
 	return h_dentry;
 }
 
@@ -196,7 +197,6 @@ int au_lkup_dentry(struct dentry *dentry, aufs_bindex_t bstart, mode_t type,
 	struct dentry *parent;
 	struct inode *inode;
 
-	parent = dget_parent(dentry);
 	err = au_test_shwh(dentry->d_sb, name);
 	if (unlikely(err))
 		goto out;
@@ -211,6 +211,7 @@ int au_lkup_dentry(struct dentry *dentry, aufs_bindex_t bstart, mode_t type,
 		au_fset_lkup(args.flags, ALLOW_NEG);
 
 	npositive = 0;
+	parent = dget_parent(dentry);
 	btail = au_dbtaildir(parent);
 	for (bindex = bstart; bindex <= btail; bindex++) {
 		struct dentry *h_parent, *h_dentry;
@@ -237,7 +238,7 @@ int au_lkup_dentry(struct dentry *dentry, aufs_bindex_t bstart, mode_t type,
 		mutex_unlock(&h_dir->i_mutex);
 		err = PTR_ERR(h_dentry);
 		if (IS_ERR(h_dentry))
-			goto out_wh;
+			goto out_parent;
 		au_fclr_lkup(args.flags, ALLOW_NEG);
 
 		if (au_dbwh(dentry) >= 0)
@@ -266,14 +267,16 @@ int au_lkup_dentry(struct dentry *dentry, aufs_bindex_t bstart, mode_t type,
 	}
 	err = npositive;
 	if (unlikely(!au_opt_test(au_mntflags(dentry->d_sb), UDBA_NONE)
-		     && au_dbstart(dentry) < 0))
-		/* both of real entry and whiteout found */
+		     && au_dbstart(dentry) < 0)) {
 		err = -EIO;
+		AuIOErr("both of real entry and whiteout found, %.*s, err %d\n",
+			AuDLNPair(dentry), err);
+	}
 
- out_wh:
-	kfree(whname.name);
- out:
+out_parent:
 	dput(parent);
+	kfree(whname.name);
+out:
 	return err;
 }
 
@@ -309,32 +312,30 @@ int au_lkup_neg(struct dentry *dentry, aufs_bindex_t bindex)
 {
 	int err;
 	struct dentry *parent, *h_parent, *h_dentry;
-	struct qstr *name;
 
-	name = &dentry->d_name;
 	parent = dget_parent(dentry);
 	h_parent = au_h_dptr(parent, bindex);
-	h_dentry = au_sio_lkup_one(name, h_parent,
+	h_dentry = au_sio_lkup_one(&dentry->d_name, h_parent,
 				   au_sbr(dentry->d_sb, bindex));
 	err = PTR_ERR(h_dentry);
 	if (IS_ERR(h_dentry))
 		goto out;
 	if (unlikely(h_dentry->d_inode)) {
 		err = -EIO;
-		AuIOErr("b%d %.*s should be negative.\n",
-			bindex, AuDLNPair(h_dentry));
+		AuIOErr("%.*s should be negative on b%d.\n",
+			AuDLNPair(h_dentry), bindex);
 		dput(h_dentry);
 		goto out;
 	}
 
+	err = 0;
 	if (bindex < au_dbstart(dentry))
 		au_set_dbstart(dentry, bindex);
 	if (au_dbend(dentry) < bindex)
 		au_set_dbend(dentry, bindex);
 	au_set_h_dptr(dentry, bindex, h_dentry);
-	err = 0;
 
- out:
+out:
 	dput(parent);
 	return err;
 }
@@ -416,7 +417,7 @@ static int au_h_verify_dentry(struct dentry *h_dentry, struct dentry *h_parent,
 		err = au_busy_or_stale();
 	dput(h_d);
 
- out:
+out:
 	AuTraceErr(err);
 	return err;
 }
@@ -427,10 +428,11 @@ int au_h_verify(struct dentry *h_dentry, unsigned int udba, struct inode *h_dir,
 	int err;
 
 	err = 0;
-	if (udba == AuOpt_UDBA_REVAL) {
+	if (udba == AuOpt_UDBA_REVAL
+	    && !au_test_fs_remote(h_dentry->d_sb)) {
 		IMustLock(h_dir);
 		err = (h_dentry->d_parent->d_inode != h_dir);
-	} else if (udba == AuOpt_UDBA_HINOTIFY)
+	} else if (udba != AuOpt_UDBA_NONE)
 		err = au_h_verify_dentry(h_dentry, h_parent, br);
 
 	return err;
@@ -438,30 +440,30 @@ int au_h_verify(struct dentry *h_dentry, unsigned int udba, struct inode *h_dir,
 
 /* ---------------------------------------------------------------------- */
 
-static void au_do_refresh_hdentry(struct au_hdentry *p, struct au_dinfo *dinfo,
-				  struct dentry *parent)
+static int au_do_refresh_hdentry(struct dentry *dentry, struct dentry *parent)
 {
-	struct dentry *h_d, *h_dp;
-	struct au_hdentry tmp, *q;
-	struct super_block *sb;
+	int err;
 	aufs_bindex_t new_bindex, bindex, bend, bwh, bdiropq;
+	struct au_hdentry tmp, *p, *q;
+	struct au_dinfo *dinfo;
+	struct super_block *sb;
 
+	DiMustWriteLock(dentry);
+
+	sb = dentry->d_sb;
+	dinfo = au_di(dentry);
 	bend = dinfo->di_bend;
 	bwh = dinfo->di_bwh;
 	bdiropq = dinfo->di_bdiropq;
+	p = dinfo->di_hdentry + dinfo->di_bstart;
 	for (bindex = dinfo->di_bstart; bindex <= bend; bindex++, p++) {
-		h_d = p->hd_dentry;
-		if (!h_d)
+		if (!p->hd_dentry)
 			continue;
 
-		h_dp = dget_parent(h_d);
-		if (h_dp == au_h_dptr(parent, bindex)) {
-			dput(h_dp);
+		new_bindex = au_br_index(sb, p->hd_id);
+		if (new_bindex == bindex)
 			continue;
-		}
 
-		new_bindex = au_find_dbindex(parent, h_dp);
-		dput(h_dp);
 		if (dinfo->di_bwh == bindex)
 			bwh = new_bindex;
 		if (dinfo->di_bdiropq == bindex)
@@ -483,7 +485,6 @@ static void au_do_refresh_hdentry(struct au_hdentry *p, struct au_dinfo *dinfo,
 		}
 	}
 
-	sb = parent->d_sb;
 	dinfo->di_bwh = -1;
 	if (bwh >= 0 && bwh <= au_sbend(sb) && au_sbr_whable(sb, bwh))
 		dinfo->di_bwh = bwh;
@@ -494,6 +495,9 @@ static void au_do_refresh_hdentry(struct au_hdentry *p, struct au_dinfo *dinfo,
 	    && au_sbr_whable(sb, bdiropq))
 		dinfo->di_bdiropq = bdiropq;
 
+	err = -EIO;
+	dinfo->di_bstart = -1;
+	dinfo->di_bend = -1;
 	bend = au_dbend(parent);
 	p = dinfo->di_hdentry;
 	for (bindex = 0; bindex <= bend; bindex++, p++)
@@ -502,59 +506,318 @@ static void au_do_refresh_hdentry(struct au_hdentry *p, struct au_dinfo *dinfo,
 			break;
 		}
 
-	p = dinfo->di_hdentry + bend;
-	for (bindex = bend; bindex >= 0; bindex--, p--)
-		if (p->hd_dentry) {
-			dinfo->di_bend = bindex;
-			break;
+	if (dinfo->di_bstart >= 0) {
+		p = dinfo->di_hdentry + bend;
+		for (bindex = bend; bindex >= 0; bindex--, p--)
+			if (p->hd_dentry) {
+				dinfo->di_bend = bindex;
+				err = 0;
+				break;
+			}
+	}
+
+	return err;
+}
+
+static void au_do_hide(struct dentry *dentry)
+{
+	struct inode *inode;
+
+	inode = dentry->d_inode;
+	if (inode) {
+		if (!S_ISDIR(inode->i_mode)) {
+			if (inode->i_nlink && !d_unhashed(dentry))
+				drop_nlink(inode);
+		} else {
+			clear_nlink(inode);
+			/* stop next lookup */
+			inode->i_flags |= S_DEAD;
 		}
+		smp_mb(); /* necessary? */
+	}
+	d_drop(dentry);
+}
+
+static int au_hide_children(struct dentry *parent)
+{
+	int err, i, j, ndentry;
+	struct au_dcsub_pages dpages;
+	struct au_dpage *dpage;
+	struct dentry *dentry;
+
+	err = au_dpages_init(&dpages, GFP_NOFS);
+	if (unlikely(err))
+		goto out;
+	err = au_dcsub_pages(&dpages, parent, NULL, NULL);
+	if (unlikely(err))
+		goto out_dpages;
+
+	/* in reverse order */
+	for (i = dpages.ndpage - 1; i >= 0; i--) {
+		dpage = dpages.dpages + i;
+		ndentry = dpage->ndentry;
+		for (j = ndentry - 1; j >= 0; j--) {
+			dentry = dpage->dentries[j];
+			if (dentry != parent)
+				au_do_hide(dentry);
+		}
+	}
+
+out_dpages:
+	au_dpages_free(&dpages);
+out:
+	return err;
+}
+
+static void au_hide(struct dentry *dentry)
+{
+	int err;
+	struct inode *inode;
+
+	AuDbgDentry(dentry);
+	inode = dentry->d_inode;
+	if (inode && S_ISDIR(inode->i_mode)) {
+		/* shrink_dcache_parent(dentry); */
+		err = au_hide_children(dentry);
+		if (unlikely(err))
+			AuIOErr("%.*s, failed hiding children, ignored %d\n",
+				AuDLNPair(dentry), err);
+	}
+	au_do_hide(dentry);
 }
 
 /*
- * returns the number of found lower positive dentries,
- * otherwise an error.
+ * By adding a dirty branch, a cached dentry may be affected in various ways.
+ *
+ * a dirty branch is added
+ * - on the top of layers
+ * - in the middle of layers
+ * - to the bottom of layers
+ *
+ * on the added branch there exists
+ * - a whiteout
+ * - a diropq
+ * - a same named entry
+ *   + exist
+ *     * negative --> positive
+ *     * positive --> positive
+ *	 - type is unchanged
+ *	 - type is changed
+ *   + doesn't exist
+ *     * negative --> negative
+ *     * positive --> negative (rejected by au_br_del() for non-dir case)
+ * - none
  */
-int au_refresh_hdentry(struct dentry *dentry, mode_t type)
+static int au_refresh_by_dinfo(struct dentry *dentry, struct au_dinfo *dinfo,
+			       struct au_dinfo *tmp)
 {
-	int npositive, err;
+	int err;
+	aufs_bindex_t bindex, bend;
+	struct {
+		struct dentry *dentry;
+		struct inode *inode;
+		mode_t mode;
+	} orig_h, tmp_h;
+	struct au_hdentry *hd;
+	struct inode *inode, *h_inode;
+	struct dentry *h_dentry;
+
+	err = 0;
+	AuDebugOn(dinfo->di_bstart < 0);
+	orig_h.dentry = dinfo->di_hdentry[dinfo->di_bstart].hd_dentry;
+	orig_h.inode = orig_h.dentry->d_inode;
+	orig_h.mode = 0;
+	if (orig_h.inode)
+		orig_h.mode = orig_h.inode->i_mode & S_IFMT;
+	memset(&tmp_h, 0, sizeof(tmp_h));
+	if (tmp->di_bstart >= 0) {
+		tmp_h.dentry = tmp->di_hdentry[tmp->di_bstart].hd_dentry;
+		tmp_h.inode = tmp_h.dentry->d_inode;
+		if (tmp_h.inode)
+			tmp_h.mode = tmp_h.inode->i_mode & S_IFMT;
+	}
+
+	inode = dentry->d_inode;
+	if (!orig_h.inode) {
+		AuDbg("nagative originally\n");
+		if (inode) {
+			au_hide(dentry);
+			goto out;
+		}
+		AuDebugOn(inode);
+		AuDebugOn(dinfo->di_bstart != dinfo->di_bend);
+		AuDebugOn(dinfo->di_bdiropq != -1);
+
+		if (!tmp_h.inode) {
+			AuDbg("negative --> negative\n");
+			/* should have only one negative lower */
+			if (tmp->di_bstart >= 0
+			    && tmp->di_bstart < dinfo->di_bstart) {
+				AuDebugOn(tmp->di_bstart != tmp->di_bend);
+				AuDebugOn(dinfo->di_bstart != dinfo->di_bend);
+				au_set_h_dptr(dentry, dinfo->di_bstart, NULL);
+				au_di_cp(dinfo, tmp);
+				hd = tmp->di_hdentry + tmp->di_bstart;
+				au_set_h_dptr(dentry, tmp->di_bstart,
+					      dget(hd->hd_dentry));
+			}
+			au_dbg_verify_dinode(dentry);
+		} else {
+			AuDbg("negative --> positive\n");
+			/*
+			 * similar to the behaviour of creating with bypassing
+			 * aufs.
+			 * unhash it in order to force an error in the
+			 * succeeding create operation.
+			 * we should not set S_DEAD here.
+			 */
+			d_drop(dentry);
+			/* au_di_swap(tmp, dinfo); */
+			au_dbg_verify_dinode(dentry);
+		}
+	} else {
+		AuDbg("positive originally\n");
+		/* inode may be NULL */
+		AuDebugOn(inode && (inode->i_mode & S_IFMT) != orig_h.mode);
+		if (!tmp_h.inode) {
+			AuDbg("positive --> negative\n");
+			/* or bypassing aufs */
+			au_hide(dentry);
+			if (tmp->di_bwh >= 0 && tmp->di_bwh <= dinfo->di_bstart)
+				dinfo->di_bwh = tmp->di_bwh;
+			if (inode)
+				err = au_refresh_hinode_self(inode);
+			au_dbg_verify_dinode(dentry);
+		} else if (orig_h.mode == tmp_h.mode) {
+			AuDbg("positive --> positive, same type\n");
+			if (!S_ISDIR(orig_h.mode)
+			    && dinfo->di_bstart > tmp->di_bstart) {
+				/*
+				 * similar to the behaviour of removing and
+				 * creating.
+				 */
+				au_hide(dentry);
+				if (inode)
+					err = au_refresh_hinode_self(inode);
+				au_dbg_verify_dinode(dentry);
+			} else {
+				/* fill empty slots */
+				if (dinfo->di_bstart > tmp->di_bstart)
+					dinfo->di_bstart = tmp->di_bstart;
+				if (dinfo->di_bend < tmp->di_bend)
+					dinfo->di_bend = tmp->di_bend;
+				dinfo->di_bwh = tmp->di_bwh;
+				dinfo->di_bdiropq = tmp->di_bdiropq;
+				hd = tmp->di_hdentry;
+				bend = dinfo->di_bend;
+				for (bindex = tmp->di_bstart; bindex <= bend;
+				     bindex++) {
+					if (au_h_dptr(dentry, bindex))
+						continue;
+					h_dentry = hd[bindex].hd_dentry;
+					if (!h_dentry)
+						continue;
+					h_inode = h_dentry->d_inode;
+					AuDebugOn(!h_inode);
+					AuDebugOn(orig_h.mode
+						  != (h_inode->i_mode
+						      & S_IFMT));
+					au_set_h_dptr(dentry, bindex,
+						      dget(h_dentry));
+				}
+				err = au_refresh_hinode(inode, dentry);
+				au_dbg_verify_dinode(dentry);
+			}
+		} else {
+			AuDbg("positive --> positive, different type\n");
+			/* similar to the behaviour of removing and creating */
+			au_hide(dentry);
+			if (inode)
+				err = au_refresh_hinode_self(inode);
+			au_dbg_verify_dinode(dentry);
+		}
+	}
+
+out:
+	return err;
+}
+
+int au_refresh_dentry(struct dentry *dentry, struct dentry *parent)
+{
+	int err, ebrange;
 	unsigned int sigen;
-	aufs_bindex_t bstart;
-	struct au_dinfo *dinfo;
+	struct au_dinfo *dinfo, *tmp;
 	struct super_block *sb;
-	struct dentry *parent;
+	struct inode *inode;
+
+	DiMustWriteLock(dentry);
+	AuDebugOn(IS_ROOT(dentry));
+	AuDebugOn(!parent->d_inode);
 
 	sb = dentry->d_sb;
-	AuDebugOn(IS_ROOT(dentry));
+	inode = dentry->d_inode;
 	sigen = au_sigen(sb);
-	parent = dget_parent(dentry);
-	AuDebugOn(au_digen(parent) != sigen
-		  || au_iigen(parent->d_inode) != sigen);
+	err = au_digen_test(parent, sigen);
+	if (unlikely(err))
+		goto out;
 
 	dinfo = au_di(dentry);
 	err = au_di_realloc(dinfo, au_sbend(sb) + 1);
-	npositive = err;
 	if (unlikely(err))
 		goto out;
-	au_do_refresh_hdentry(dinfo->di_hdentry + dinfo->di_bstart, dinfo,
-			      parent);
+	ebrange = au_dbrange_test(dentry);
+	if (!ebrange)
+		ebrange = au_do_refresh_hdentry(dentry, parent);
 
-	npositive = 0;
-	bstart = au_dbstart(parent);
-	if (type != S_IFDIR && dinfo->di_bstart == bstart)
-		goto out_dgen; /* success */
-
-	npositive = au_lkup_dentry(dentry, bstart, type, /*nd*/NULL);
-	if (npositive < 0)
+	if (d_unhashed(dentry) || ebrange) {
+		AuDebugOn(au_dbstart(dentry) < 0 && au_dbend(dentry) >= 0);
+		if (inode)
+			err = au_refresh_hinode_self(inode);
+		au_dbg_verify_dinode(dentry);
+		if (!err)
+			goto out_dgen; /* success */
 		goto out;
-	if (dinfo->di_bwh >= 0 && dinfo->di_bwh <= dinfo->di_bstart)
-		d_drop(dentry);
+	}
 
- out_dgen:
+	/* temporary dinfo */
+	AuDbgDentry(dentry);
+	err = -ENOMEM;
+	tmp = au_di_alloc(sb, AuLsc_DI_TMP);
+	if (unlikely(!tmp))
+		goto out;
+	au_di_swap(tmp, dinfo);
+	/* returns the number of positive dentries */
+	/*
+	 * if current working dir is removed, it returns an error.
+	 * but the dentry is legal.
+	 */
+	err = au_lkup_dentry(dentry, /*bstart*/0, /*type*/0, /*nd*/NULL);
+	AuDbgDentry(dentry);
+	au_di_swap(tmp, dinfo);
+	if (err == -ENOENT)
+		err = 0;
+	if (err >= 0) {
+		/* compare/refresh by dinfo */
+		AuDbgDentry(dentry);
+		err = au_refresh_by_dinfo(dentry, dinfo, tmp);
+		au_dbg_verify_dinode(dentry);
+		AuTraceErr(err);
+	}
+	au_rw_write_unlock(&tmp->di_rwsem);
+	au_di_free(tmp);
+	if (unlikely(err))
+		goto out;
+
+out_dgen:
 	au_update_digen(dentry);
- out:
-	dput(parent);
-	AuTraceErr(npositive);
-	return npositive;
+out:
+	if (unlikely(err && !(dentry->d_flags & DCACHE_NFSFS_RENAMED))) {
+		AuIOErr("failed refreshing %.*s, %d\n",
+			AuDLNPair(dentry), err);
+		AuDbgDentry(dentry);
+	}
+	AuTraceErr(err);
+	return err;
 }
 
 static noinline_for_stack
@@ -565,11 +828,9 @@ int au_do_h_d_reval(struct dentry *h_dentry, struct nameidata *nd,
 	int (*reval)(struct dentry *, struct nameidata *);
 
 	err = 0;
-	reval = NULL;
-	if (h_dentry->d_op)
-		reval = h_dentry->d_op->d_revalidate;
-	if (!reval)
+	if (!(h_dentry->d_flags & DCACHE_OP_REVALIDATE))
 		goto out;
+	reval = h_dentry->d_op->d_revalidate;
 
 	AuDbg("b%d\n", bindex);
 	if (au_test_fs_null_nd(h_dentry->d_sb))
@@ -601,7 +862,7 @@ int au_do_h_d_reval(struct dentry *h_dentry, struct nameidata *nd,
 	else if (!valid)
 		err = -EINVAL;
 
- out:
+out:
 	AuTraceErr(err);
 	return err;
 }
@@ -614,14 +875,13 @@ static int h_d_revalidate(struct dentry *dentry, struct inode *inode,
 	umode_t mode, h_mode;
 	aufs_bindex_t bindex, btail, bstart, ibs, ibe;
 	unsigned char plus, unhashed, is_root, h_plus;
-	struct inode *first, *h_inode, *h_cached_inode;
+	struct inode *h_inode, *h_cached_inode;
 	struct dentry *h_dentry;
 	struct qstr *name, *h_name;
 
 	err = 0;
 	plus = 0;
 	mode = 0;
-	first = NULL;
 	ibs = -1;
 	ibe = -1;
 	unhashed = !!d_unhashed(dentry);
@@ -629,8 +889,9 @@ static int h_d_revalidate(struct dentry *dentry, struct inode *inode,
 	name = &dentry->d_name;
 
 	/*
-	 * Theoretically, REVAL test should be unnecessary in case of INOTIFY.
-	 * But inotify doesn't fire some necessary events,
+	 * Theoretically, REVAL test should be unnecessary in case of
+	 * {FS,I}NOTIFY.
+	 * But {fs,i}notify doesn't fire some necessary events,
 	 *	IN_ATTRIB for atime/nlink/pageio
 	 *	IN_DELETE for NFS dentry
 	 * Let's do REVAL test too.
@@ -638,7 +899,6 @@ static int h_d_revalidate(struct dentry *dentry, struct inode *inode,
 	if (do_udba && inode) {
 		mode = (inode->i_mode & S_IFMT);
 		plus = (inode->i_nlink > 0);
-		first = au_h_iptr(inode, au_ibstart(inode));
 		ibs = au_ibstart(inode);
 		ibe = au_ibend(inode);
 	}
@@ -653,6 +913,7 @@ static int h_d_revalidate(struct dentry *dentry, struct inode *inode,
 			continue;
 
 		AuDbg("b%d, %.*s\n", bindex, AuDLNPair(h_dentry));
+		spin_lock(&h_dentry->d_lock);
 		h_name = &h_dentry->d_name;
 		if (unlikely(do_udba
 			     && !is_root
@@ -663,8 +924,10 @@ static int h_d_revalidate(struct dentry *dentry, struct inode *inode,
 			AuDbg("unhash 0x%x 0x%x, %.*s %.*s\n",
 				  unhashed, d_unhashed(h_dentry),
 				  AuDLNPair(dentry), AuDLNPair(h_dentry));
+			spin_unlock(&h_dentry->d_lock);
 			goto err;
 		}
+		spin_unlock(&h_dentry->d_lock);
 
 		err = au_do_h_d_reval(h_dentry, nd, dentry, bindex);
 		if (unlikely(err))
@@ -704,29 +967,23 @@ static int h_d_revalidate(struct dentry *dentry, struct inode *inode,
 	return err;
 }
 
+/* todo: consolidate with do_refresh() and au_reval_for_attr() */
 static int simple_reval_dpath(struct dentry *dentry, unsigned int sigen)
 {
 	int err;
 	struct dentry *parent;
-	struct inode *inode;
 
-	inode = dentry->d_inode;
-	if (au_digen(dentry) == sigen && au_iigen(inode) == sigen)
+	if (!au_digen_test(dentry, sigen))
 		return 0;
 
 	parent = dget_parent(dentry);
 	di_read_lock_parent(parent, AuLock_IR);
-	AuDebugOn(au_digen(parent) != sigen
-		  || au_iigen(parent->d_inode) != sigen);
+	AuDebugOn(au_digen_test(parent, sigen));
 	au_dbg_verify_gen(parent, sigen);
-
-	/* returns a number of positive dentries */
-	err = au_refresh_hdentry(dentry, inode->i_mode & S_IFMT);
-	if (err >= 0)
-		err = au_refresh_hinode(inode, dentry);
-
+	err = au_refresh_dentry(dentry, parent);
 	di_read_unlock(parent, AuLock_IR);
 	dput(parent);
+	AuTraceErr(err);
 	return err;
 }
 
@@ -736,36 +993,35 @@ int au_reval_dpath(struct dentry *dentry, unsigned int sigen)
 	struct dentry *d, *parent;
 	struct inode *inode;
 
-	if (!au_ftest_si(au_sbi(dentry->d_sb), FAILED_REFRESH_DIRS))
+	if (!au_ftest_si(au_sbi(dentry->d_sb), FAILED_REFRESH_DIR))
 		return simple_reval_dpath(dentry, sigen);
 
 	/* slow loop, keep it simple and stupid */
 	/* cf: au_cpup_dirs() */
 	err = 0;
 	parent = NULL;
-	while (au_digen(dentry) != sigen
-	       || au_iigen(dentry->d_inode) != sigen) {
+	while (au_digen_test(dentry, sigen)) {
 		d = dentry;
 		while (1) {
 			dput(parent);
 			parent = dget_parent(d);
-			if (au_digen(parent) == sigen
-			    && au_iigen(parent->d_inode) == sigen)
+			if (!au_digen_test(parent, sigen))
 				break;
 			d = parent;
 		}
 
 		inode = d->d_inode;
 		if (d != dentry)
-			di_write_lock_child(d);
+			di_write_lock_child2(d);
 
 		/* someone might update our dentry while we were sleeping */
-		if (au_digen(d) != sigen || au_iigen(d->d_inode) != sigen) {
+		if (au_digen_test(d, sigen)) {
+			/*
+			 * todo: consolidate with simple_reval_dpath(),
+			 * do_refresh() and au_reval_for_attr().
+			 */
 			di_read_lock_parent(parent, AuLock_IR);
-			/* returns a number of positive dentries */
-			err = au_refresh_hdentry(d, inode->i_mode & S_IFMT);
-			if (err >= 0)
-				err = au_refresh_hinode(inode, d);
+			err = au_refresh_dentry(d, parent);
 			di_read_unlock(parent, AuLock_IR);
 		}
 
@@ -790,86 +1046,95 @@ static int aufs_d_revalidate(struct dentry *dentry, struct nameidata *nd)
 	struct super_block *sb;
 	struct inode *inode;
 
-	err = -EINVAL;
-	sb = dentry->d_sb;
+	/* todo: support rcu-walk? */
+	if (nd && (nd->flags & LOOKUP_RCU))
+		return -ECHILD;
+
+	valid = 0;
+	if (unlikely(!au_di(dentry)))
+		goto out;
+
 	inode = dentry->d_inode;
-	aufs_read_lock(dentry, AuLock_FLUSH | AuLock_DW);
-	sigen = au_sigen(sb);
-	if (au_digen(dentry) != sigen) {
-		AuDebugOn(IS_ROOT(dentry));
-		if (inode)
-			err = au_reval_dpath(dentry, sigen);
-		if (unlikely(err))
-			goto out_dgrade;
-		AuDebugOn(au_digen(dentry) != sigen);
+	if (inode && is_bad_inode(inode))
+		goto out;
+
+	valid = 1;
+	sb = dentry->d_sb;
+	/*
+	 * todo: very ugly
+	 * i_mutex of parent dir may be held,
+	 * but we should not return 'invalid' due to busy.
+	 */
+	err = aufs_read_lock(dentry, AuLock_FLUSH | AuLock_DW | AuLock_NOPLM);
+	if (unlikely(err)) {
+		valid = err;
+		AuTraceErr(err);
+		goto out;
 	}
-	if (inode && au_iigen(inode) != sigen) {
+	if (unlikely(au_dbrange_test(dentry))) {
+		err = -EINVAL;
+		AuTraceErr(err);
+		goto out_dgrade;
+	}
+
+	sigen = au_sigen(sb);
+	if (au_digen_test(dentry, sigen)) {
 		AuDebugOn(IS_ROOT(dentry));
-		err = au_refresh_hinode(inode, dentry);
-		if (unlikely(err))
+		err = au_reval_dpath(dentry, sigen);
+		if (unlikely(err)) {
+			AuTraceErr(err);
 			goto out_dgrade;
-		AuDebugOn(au_iigen(inode) != sigen);
+		}
 	}
 	di_downgrade_lock(dentry, AuLock_IR);
 
-	AuDebugOn(au_digen(dentry) != sigen);
-	AuDebugOn(inode && au_iigen(inode) != sigen);
 	err = -EINVAL;
+	if (inode && (IS_DEADDIR(inode) || !inode->i_nlink))
+		goto out_inval;
+
 	do_udba = !au_opt_test(au_mntflags(sb), UDBA_NONE);
 	if (do_udba && inode) {
 		aufs_bindex_t bstart = au_ibstart(inode);
+		struct inode *h_inode;
 
-		if (bstart >= 0
-		    && au_test_higen(inode, au_h_iptr(inode, bstart)))
-			goto out;
+		if (bstart >= 0) {
+			h_inode = au_h_iptr(inode, bstart);
+			if (h_inode && au_test_higen(inode, h_inode))
+				goto out_inval;
+		}
 	}
 
 	err = h_d_revalidate(dentry, inode, nd, do_udba);
-	if (unlikely(!err && do_udba && au_dbstart(dentry) < 0))
-		/* both of real entry and whiteout found */
+	if (unlikely(!err && do_udba && au_dbstart(dentry) < 0)) {
 		err = -EIO;
-	goto out;
+		AuDbg("both of real entry and whiteout found, %.*s, err %d\n",
+		      AuDLNPair(dentry), err);
+	}
+	goto out_inval;
 
- out_dgrade:
+out_dgrade:
 	di_downgrade_lock(dentry, AuLock_IR);
- out:
+out_inval:
 	aufs_read_unlock(dentry, AuLock_IR);
 	AuTraceErr(err);
 	valid = !err;
-	if (!valid)
-		AuDbg("%.*s invalid\n", AuDLNPair(dentry));
+out:
+	if (!valid) {
+		AuDbg("%.*s invalid, %d\n", AuDLNPair(dentry), valid);
+		d_drop(dentry);
+	}
 	return valid;
 }
 
 static void aufs_d_release(struct dentry *dentry)
 {
-	struct au_dinfo *dinfo;
-	aufs_bindex_t bend, bindex;
-
-	dinfo = dentry->d_fsdata;
-	if (!dinfo)
-		return;
-
-	/* dentry may not be revalidated */
-	bindex = dinfo->di_bstart;
-	if (bindex >= 0) {
-		struct au_hdentry *p;
-
-		bend = dinfo->di_bend;
-		p = dinfo->di_hdentry + bindex;
-		while (bindex++ <= bend) {
-			if (p->hd_dentry)
-				au_hdput(p);
-			p++;
-		}
+	if (au_di(dentry)) {
+		au_di_fin(dentry);
+		au_hn_di_reinit(dentry);
 	}
-	kfree(dinfo->di_hdentry);
-	au_rwsem_destroy(&dinfo->di_rwsem);
-	au_cache_free_dinfo(dinfo);
-	au_hin_di_reinit(dentry);
 }
 
-struct dentry_operations aufs_dop = {
+const struct dentry_operations aufs_dop = {
 	.d_revalidate	= aufs_d_revalidate,
 	.d_release	= aufs_d_release
 };
